using System;
using System.Collections;
using System.ComponentModel;
using System.IO;
using System.Net;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;

using Microsoft.Win32;

namespace archiveLP
{
   public struct SAcookies
   {
      public int bbuserid;
      public string bbpassword;
      public string browserName;
   }

   public partial class archiveLP : Form
   {
      // SQLite DLL imports
      [DllImport("sqlite3.dll")]
      unsafe public static extern int sqlite3_open_v2(IntPtr filename, IntPtr* ppDb, int flags, IntPtr zVfs);
      [DllImport("sqlite3.dll")]
      unsafe public static extern int sqlite3_close(IntPtr pDb);
      [DllImport("sqlite3.dll")]
      unsafe public static extern int sqlite3_get_table(IntPtr pDb, IntPtr sql,
         IntPtr* pResArr, int* pnRow, int* pnColumn, IntPtr* pErrMsg);
      [DllImport("sqlite3.dll")]
      unsafe public static extern void sqlite3_free(IntPtr mem);
      [DllImport("sqlite3.dll")]
      unsafe public static extern void sqlite3_free_table(IntPtr pResArr);
      public static int SQLITE_OK = 0;
      public static int SQLITE_OPEN_READONLY = 1;

      // Cookie settings
      internal bool useManualCook = false;
      internal string manualCookUser = "123456";
      internal string manualCookPwd = "5a244235d22877ecf851620402e13129";

      internal bool autoCookIE = true;
      internal bool autoCookFF2 = true;
      internal bool autoCookFF3 = true;
      internal bool autoCookOpera = true;

      // Lists settings
      internal string[] videoSiteNames;
      internal string[] videoSiteURLs;
      internal string[] chapterWords;

      // Fixed settings
      static string cssURL = "/LetsPlay/myStyle.css";
      internal static string[] videoSiteNamesDefault = { "Dailymotion", "YouTube", "Google", "FileFront", "Vimeo", "blip.tv", "Viddler" };
      internal static string[] videoSiteURLsDefault = { "dailymotion.com", "youtube.com", "video.google.[^/]+", "filefront.com", "vimeo.com", "blip.tv", "viddler.com" };
      internal static string[] chapterWordsDefault = { "Chapter ", "Update ", "Episode " };

      // For inputForm's benefit
      internal static string getInput = "";

      // Avoidably global variables are the best
      StreamWriter logFile;
      logForm logger;

      // I'm a terrible person
      public struct getPostStuff
      {
         public StreamReader threadFile;
         public int postNum;
         public bool amChapterTitleChecking;
         public bool useAutoChapters;
         public int chapTitlesCount;
         public ArrayList videoURLs;
         public ArrayList videoTitles;
         public int vidColumns;
         public int imgCount;
         public Hashtable smilies;
         public int updCount;
         public bool gotUpdate;
         public ArrayList chapterTitles;
         public string outDirToo;
         public long imgSize;
         public long vidSize;
         public int postCountToo;
         public string errors;
         public ArrayList failedImages;
         public ArrayList failedVideos;
         public ArrayList postStats;
         public string vlpContent;
      }

      public struct failedImage
      {
         public string imagePath;
         public string imageURL;
         public string imageInfo;
      }
      public struct failedVideo
      {
         public string videoPath;
         public string videoURL;
         public string videoInfo;
      }

      [STAThreadAttribute]
      public static void Main()
      {
         archiveLP aLP = new archiveLP();

         loadSettings(aLP);
         aLP.fileSaver.InitialDirectory = System.Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
         aLP.fileOpener.InitialDirectory = aLP.fileSaver.InitialDirectory;
         aLP.FormClosed += new FormClosedEventHandler(exitApp);

         Application.Run(aLP);
      }

      static bool regGetBool(string regVal, RegistryKey regKey, bool defVal)
      {
         int bVal;

         try { bVal = (int)regKey.GetValue(regVal, -1); }
         catch(InvalidCastException ) { bVal = -1; }

         if (bVal == -1) { return defVal; }
         else if (bVal == 1) { return true; }
         else { return false; }
      }
      static void regSetBool(string regVal, RegistryKey regKey, bool val)
      {
         if (val) { regKey.SetValue(regVal, 1); }
         else { regKey.SetValue(regVal, 0); }
      }

      static void loadSettings(archiveLP thisForm)
      {
         // Load cookie settings from registry
         RegistryKey aLPConf = Registry.CurrentUser.OpenSubKey("Software\\Polsy\\archiveLP");
         if (aLPConf != null) // probably doesn't exist yet
         {
            // Cookie settings
            thisForm.useManualCook = regGetBool("useManualCook", aLPConf, false);

            try { thisForm.manualCookUser = (string)aLPConf.GetValue("manualCookUser", "123546"); }
            catch { thisForm.manualCookUser = "123456"; }
            if (thisForm.manualCookUser.Length > 7)
               thisForm.manualCookUser = thisForm.manualCookUser.Substring(0, 7);

            try { thisForm.manualCookPwd = (string)aLPConf.GetValue("manualCookPwd", "5a244235d22877ecf851620402e13129"); }
            catch { thisForm.manualCookPwd = "5a244235d22877ecf851620402e13129"; }
            if (thisForm.manualCookPwd.Length > 32)
               thisForm.manualCookPwd = thisForm.manualCookPwd.Substring(0, 32);

            thisForm.autoCookIE = regGetBool("autoCookIE", aLPConf, true);
            thisForm.autoCookFF2 = regGetBool("autoCookFF2", aLPConf, true);
            thisForm.autoCookFF3 = regGetBool("autoCookFF3", aLPConf, true);
            thisForm.autoCookOpera = regGetBool("autoCookOpera", aLPConf, true);

            // List settings
            try { thisForm.videoSiteNames = (string[])aLPConf.GetValue("videoSiteNames", videoSiteNamesDefault); }
            catch { thisForm.videoSiteNames = videoSiteNamesDefault; }

            try { thisForm.videoSiteURLs = (string[])aLPConf.GetValue("videoSiteURLs", videoSiteURLsDefault); }
            catch { thisForm.videoSiteURLs = videoSiteURLsDefault; }

            try { thisForm.chapterWords = (string[])aLPConf.GetValue("chapterWords", chapterWordsDefault); }
            catch { thisForm.chapterWords = chapterWordsDefault; }

            aLPConf.Close();
         }
      }

      static void saveSettings(archiveLP thisForm)
      {
         // Save cookie settings to the registry
         RegistryKey aLPConf = Registry.CurrentUser.CreateSubKey("Software\\Polsy\\archiveLP");
         if (aLPConf != null) // if it failed, just do nothing
         {
            regSetBool("useManualCook", aLPConf, thisForm.useManualCook);
            aLPConf.SetValue("manualCookUser", thisForm.manualCookUser);
            aLPConf.SetValue("manualCookPwd", thisForm.manualCookPwd);

            regSetBool("autoCookIE", aLPConf, thisForm.autoCookIE);
            regSetBool("autoCookFF2", aLPConf, thisForm.autoCookFF2);
            regSetBool("autoCookFF3", aLPConf, thisForm.autoCookFF3);
            regSetBool("autoCookOpera", aLPConf, thisForm.autoCookOpera);

            aLPConf.SetValue("videoSiteNames", thisForm.videoSiteNames);
            aLPConf.SetValue("videoSiteURLs", thisForm.videoSiteURLs);
            aLPConf.SetValue("chapterWords", thisForm.chapterWords);

            aLPConf.Close();
         }
      }

      public archiveLP()
      {
         InitializeComponent();
      }

      private void FileExit_Click(object sender, EventArgs e)
      {
         Application.Exit();
      }

      static void exitApp(object sender, FormClosedEventArgs e)
      {
         saveSettings((archiveLP)sender);
      }

      private void btnPickOutFlr_Click(object sender, EventArgs e)
      {
         folderPicker.Description = "Select output folder for LP";
         folderPicker.ShowDialog();
         txtOutFlr.Text = folderPicker.SelectedPath;
      }

      private void OptionsCookies_Click(object sender, EventArgs e)
      {
         cookieForm cF = new cookieForm(this);
         cF.ShowDialog();
      }

      internal SAcookies getCookiesIE()
      {
         // Find *forums.somethingawful* files in the Cookies directory, and
         // search them for the 'bbuserid' and 'bbpassword' lines

         SAcookies result = new SAcookies();
         result.bbuserid = -1; result.bbpassword = ""; result.browserName = "Internet Explorer";

         string cookiePath = Environment.GetFolderPath(Environment.SpecialFolder.Cookies);
         // Need to search subdirectories for Cookies\Low in Vista
         string[] saCookFiles = Directory.GetFiles(cookiePath, "*forums.somethingawful*", SearchOption.AllDirectories);

         foreach (string cFile in saCookFiles)
         {
            StreamReader cStream = new StreamReader(cFile);
            string line;

            while ((line = cStream.ReadLine()) != null)
            {
               if (line == "bbuserid")
               {
                  result.bbuserid = Convert.ToInt32(cStream.ReadLine());
               }
               if (line == "bbpassword")
               {
                  result.bbpassword = cStream.ReadLine();
               }
            }

            if (result.bbuserid != -1 && result.bbpassword != "")
            {
               cStream.Close();
               break;
            }
            cStream.Close();
         }

         return result;
      }

      internal SAcookies getCookiesFF2()
      {
         // Get the first profile in profiles.ini, then open the cookies.txt for that profile
         // Search it for a 'forums.somethingawful.com' line with 'bbuserid' or 'bbpassword'

         Regex cookieRE = new Regex(@"^\.?forums\.somethingawful\.com\s+\S+\s+\S+\s+\S+\s+\S+\s+(bb\S+)\s+(\S+)$");
         SAcookies result = new SAcookies();
         result.bbuserid = -1; result.bbpassword = ""; result.browserName = "Firefox 2";

         string ff2Dir = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) +
            "\\Mozilla\\Firefox\\";
         if (!File.Exists(ff2Dir + "profiles.ini"))
         {
            // Firefox not installed/misinstalled
            return result;
         }

         StreamReader cStream = new StreamReader(ff2Dir + "profiles.ini");

         string line, profDir = "";
         while ((line = cStream.ReadLine()) != null)
         {
            if (line.StartsWith("Path="))
            {
               profDir = line.Substring(5) + "\\";
            }
         }

         if (profDir != "")
         {
            if (!File.Exists(ff2Dir + profDir + "cookies.txt"))
            {
               // No cookies.txt - FF3?
               return result;
            }

            cStream = new StreamReader(ff2Dir + profDir + "cookies.txt");

            while ((line = cStream.ReadLine()) != null)
            {
               Match cM = cookieRE.Match(line);
               if (cM.Success)
               {
                  if (cM.Groups[1].Value == "bbuserid")
                  {
                     result.bbuserid = Convert.ToInt32(cM.Groups[2].Value);
                  }
                  else if (cM.Groups[1].Value == "bbpassword")
                  {
                     result.bbpassword = cM.Groups[2].Value;
                  }
               }
            }
            cStream.Close();
         }

         return result;
      }

      unsafe internal SAcookies getCookiesFF3()
      {
         // Same as FF2, except the file's cookies.sqlite now
         // Query the cookies database for the values

         SAcookies result = new SAcookies();
         result.bbuserid = -1; result.bbpassword = ""; result.browserName = "Firefox 3";

         string ff3Dir = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) +
            "\\Mozilla\\Firefox\\";
         if (!File.Exists(ff3Dir + "profiles.ini"))
         {
            // Firefox not installed/misinstalled
            return result;
         }

         StreamReader cStream = new StreamReader(ff3Dir + "profiles.ini");
         string line, profDir = "";
         while ((line = cStream.ReadLine()) != null)
         {
            if (line.StartsWith("Path="))
            {
               profDir = line.Substring(5);
            }
         }

         if (profDir != "")
         {
            IntPtr cookieDB = new IntPtr();
            IntPtr dbName = Marshal.StringToHGlobalAnsi(ff3Dir + profDir + "\\cookies.sqlite");
            int sqlResult;

            try
            {
               sqlResult = sqlite3_open_v2(dbName, &cookieDB, SQLITE_OPEN_READONLY, (IntPtr)null);
            }
            catch (DllNotFoundException )
            {
               MessageBox.Show("Firefox 3 cookies will not be available without the 'sqlite3.dll' file " +
                  "(which should be available from the same place you got this)", "LP Archiver");
               return result;
            }

            Marshal.FreeHGlobal(dbName);

            if (sqlResult == SQLITE_OK)
            {
               IntPtr sqlQry = Marshal.StringToHGlobalAnsi("SELECT name,value FROM moz_cookies " +
                  "WHERE name LIKE 'bb%' AND host='forums.somethingawful.com';");
               int numRows, numCols;
               IntPtr errMsg = new IntPtr();
               IntPtr sqlOutput = new IntPtr();

               sqlResult = sqlite3_get_table(cookieDB, sqlQry, &sqlOutput, &numRows, &numCols, &errMsg);
               if (sqlResult == SQLITE_OK && numRows != 0)
               {
                  IntPtr *sqlOutPtr = (IntPtr*)sqlOutput.ToPointer();
                  string name, value;

                  for (int r = 1; r <= numRows; r++) // first row is the headings, so there's an 'extra'
                  {
                     name = Marshal.PtrToStringAnsi(*(sqlOutPtr + (r * numCols)));
                     value = Marshal.PtrToStringAnsi(*(sqlOutPtr + ((r * numCols) + 1)));
                     if (name == "bbuserid")
                     {
                        result.bbuserid = Convert.ToInt32(value);
                     }
                     else if (name == "bbpassword")
                     {
                        result.bbpassword = value;
                     }
                  }
                  sqlite3_free_table(sqlOutput);
               }
               else if(sqlResult != SQLITE_OK)
               {
                  if (errMsg != null)
                  {
                     MessageBox.Show("FF3 Database error: " + Marshal.PtrToStringAnsi((IntPtr)errMsg.ToPointer()));
                     sqlite3_free(errMsg);
                  }
               }

               sqlite3_close(cookieDB);
            }
         }

         return result;
      }

      internal SAcookies getCookiesOpera()
      {
         // Process the cookies file, find 'forums.somethingawful.com' line with 'bbuserid' or 'bbpassword'
         // See http://www.opera.com/docs/fileformats/

         SAcookies result = new SAcookies();
         result.bbuserid = -1; result.bbpassword = ""; result.browserName = "Opera";

         string cookieFile = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) +
            "\\Opera\\Opera\\profile\\cookies4.dat";

         if (!File.Exists(cookieFile))
         {
            // Opera probably not installed
            return result;
         }

         FileStream cStream = new FileStream(cookieFile, FileMode.Open, FileAccess.Read);
         // ASCII encoding is a lazyish way to avoid PeekChar() trying to read more than one byte
         BinaryReader cFile = new BinaryReader(cStream, Encoding.ASCII);

         // Thanks for not having a 'read in network order' switch, BinaryReader
         Int32 fileVerNum = IPAddress.NetworkToHostOrder(cFile.ReadInt32()); // file format version
         Int32 appVerNum = IPAddress.NetworkToHostOrder(cFile.ReadInt32());  // application version
         Int16 tagLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());     // number of bytes in ID tag (assume 1)
         Int16 lenLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());     // number of bytes in length field (assume 2)

         int recordTag = 0;
         ArrayList domainList = new ArrayList();
         while (cFile.PeekChar() != -1 && recordTag != 0x84) // Apparently this is the only way to detect EOF
         {
            recordTag = cFile.ReadByte();
            if (recordTag == 1)
            {
               processOperaDomain(ref cFile, ref domainList, ref result);
            }
            else if (recordTag != 0x84) // the 'top' domain is finished if 0x84
            {
               //MessageBox.Show("RT: " + Convert.ToInt32(recordTag));
               break;
            }
         }
         cFile.Close(); cStream.Close();

         return result;
      }

      void processOperaDomain(ref BinaryReader cFile, ref ArrayList domainList, ref SAcookies result)
      {
         Int16 domRecLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
         long domEndPos = cFile.BaseStream.Position + domRecLen;

         int domTag;
         while (cFile.BaseStream.Position < domEndPos)
         {
            domTag = cFile.ReadByte();
            switch (domTag)
            {
               case 0x1E:
                  Int16 strLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
                  char[] domName = cFile.ReadChars(strLen);
                  domainList.Add(new String(domName));
                  break;
               case 0x1F: // Cookie filtering
               case 0x21: // Broken cookie handling
               case 0x25: // Third-party cookie filtering
               case 0x26: // I wonder what this is? It's not documented
                  // Apparently a "int8" is not stored in one byte. It's probably four.
                  Int16 dataLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
                  cFile.ReadBytes(dataLen);
                  break;
               default:
                  //MessageBox.Show("DT: " + Convert.ToInt32(domTag));
                  break;
            }
         }

         // path components, cookies, and subdomains go here
         int domPartTag;
         do
         {
            domPartTag = cFile.ReadByte();
            switch (domPartTag)
            {
               case 0x01: // subdomain
                  processOperaDomain(ref cFile, ref domainList, ref result);
                  break;
               case 0x02: // cookie path. I don't care about these at all.
                  Int16 pLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
                  cFile.ReadBytes(pLen);
                  break;
               case 0x03: // a cookie
                  processOperaCookie(ref cFile, domainList, ref result);
                  break;
               case 0x84:
                  // end of domain. exit the loop.
                  break;
               case 0x85:
                  // end of a path component. do nothing.
                  break;
               default:
                  //MessageBox.Show("DPT: " + Convert.ToInt32(domPartTag));
                  break;
            }
         } while (domPartTag != 0x84);

         domainList.RemoveAt(domainList.Count - 1);
      }

      void processOperaCookie(ref BinaryReader cFile, ArrayList domainList, ref SAcookies result)
      {
         Int16 cookRecLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
         long cookEndPos = cFile.BaseStream.Position + cookRecLen;

         int cookTag;
         string thisName = "", thisValue = ""; Int16 strLen;
         while (cFile.BaseStream.Position < cookEndPos)
         {
            cookTag = cFile.ReadByte();
            switch (cookTag)
            {
               case 0x10: // Cookie name
                  strLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
                  thisName = new String(cFile.ReadChars(strLen));
                  break;
               case 0x11: // Cookie value
                  strLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
                  thisValue = new String(cFile.ReadChars(strLen));
                  break;
               default:
                  if (cookTag > 0x7F)
                  {
                     // do nothing, it's a one-byte flag
                  }
                  else // read the item size and skip over it
                  {
                     Int16 itemLen = IPAddress.NetworkToHostOrder(cFile.ReadInt16());
                     cFile.ReadBytes(itemLen);
                  }
                  break;
            }
         }

         if (thisName == "bbuserid" || thisName == "bbpassword")
            if ((string)domainList[0] == "com" && (string)domainList[1] == "somethingawful" && (string)domainList[2] == "forums")
               if (thisName == "bbuserid")
                  result.bbuserid = Convert.ToInt32(thisValue);
               else
                  result.bbpassword = thisValue;
      }

      private void OptionsLists_Click(object sender, EventArgs e)
      {
         listsForm lF = new listsForm(this);
         lF.ShowDialog();
      }

      private void writeCheck(CheckBox c, StreamWriter sw)
      {
         if (c.Checked)
            sw.WriteLine("1");
         else
            sw.WriteLine("0");
      }
      private void FileSave_Click(object sender, EventArgs e)
      {
         if (fileSaver.ShowDialog() == DialogResult.OK)
         {
            try
            {
               StreamWriter saveFile = new StreamWriter(fileSaver.FileName);
               saveFile.WriteLine(txtURL.Text);
               saveFile.WriteLine(txtLPName.Text);
               saveFile.WriteLine(txtCollName.Text);
               saveFile.WriteLine(txtOutFlr.Text);

               if (optSSLP.Checked)
                  saveFile.WriteLine("0");
               else
                  saveFile.WriteLine("1");

               writeCheck(chkDownloadFiles, saveFile);
               writeCheck(chkAutoChapterNaming, saveFile);
               saveFile.WriteLine(txtMinImages.Text);
               writeCheck(chkKeepVideoOnlyPosts, saveFile);
               writeCheck(chkAuthorPostsOnly, saveFile);
               saveFile.WriteLine(txtAutoChapThresh.Text);

               saveFile.Close();

               fileOpener.FileName = fileSaver.FileName;
            }
            catch (Exception ex)
            {
               MessageBox.Show("Error saving file: " + ex.Message);
            }
         }
      }

      private void readCheck(CheckBox c, StreamReader sr)
      {
         string tmpy = sr.ReadLine();

         if (tmpy == "1")
            c.Checked = true;
         else
            c.Checked = false;
      }
      private void FileLoad_Click(object sender, EventArgs e)
      {
         if (fileOpener.ShowDialog() == DialogResult.OK)
         {
            try
            {
               string tmpy;

               StreamReader loadFile = new StreamReader(fileOpener.FileName);
               txtURL.Text = loadFile.ReadLine();
               txtLPName.Text = loadFile.ReadLine();
               txtCollName.Text = loadFile.ReadLine();
               txtOutFlr.Text = loadFile.ReadLine();

               tmpy = loadFile.ReadLine();
               if (tmpy == "0")
                  optSSLP.Checked = true;
               else
                  optVLP.Checked = true;

               readCheck(chkDownloadFiles, loadFile);
               readCheck(chkAutoChapterNaming, loadFile);
               txtMinImages.Text = loadFile.ReadLine();
               readCheck(chkKeepVideoOnlyPosts, loadFile);
               readCheck(chkAuthorPostsOnly, loadFile);
               txtAutoChapThresh.Text = loadFile.ReadLine();

               loadFile.Close();

               fileSaver.FileName = fileOpener.FileName;
            }
            catch (Exception ex)
            {
               MessageBox.Show("Error loading file: " + ex.Message);
            }
         }
      }

      private void doLog(string what)
      {
         logger.logList.Items.Add(what);
         // Scroll to the bottom
         logger.logList.SelectedIndex = logger.logList.Items.Count - 1;
         // leaving the item selected is ugly. this is lame too, but it's still prettier.
         logger.logList.SelectedIndex = -1;
         Application.DoEvents();
         logFile.WriteLine(what);
      }

      private void btnGo_Click(object sender, EventArgs e)
      {
         // Pre-start validation
         bool proceed = true;

         // Are things that should be numbers actually numbers?
         Regex isNum = new Regex(@"^\d+$");

         if(! isNum.Match(txtMinImages.Text).Success) {
            MessageBox.Show("'Minimum number of images' must be a number", "LP Archiver");
            proceed = false;
         }
         if(! isNum.Match(txtAutoChapThresh.Text).Success) {
            MessageBox.Show("Automatic chapter naming threshold must be a number", "LP Archiver");
            proceed = false;
         }
         if (useManualCook && !isNum.Match(manualCookUser).Success)
         {
            MessageBox.Show("Manual cookie user ID must be a number", "LP Archiver");
            proceed = false;
         }

         // Output directory available?
         string outDir = txtOutFlr.Text + "\\";

         if (outDir == "\\")
         {
            MessageBox.Show("Output folder is not set", "LP Archiver");
            proceed = false;
         }
         else
         {
            if (!Directory.Exists(outDir))
            {
               try { Directory.CreateDirectory(outDir); }
               catch (Exception ex)
               {
                  MessageBox.Show("Could not create output directory: " + ex.Message, "LP Archiver");
                  proceed = false;
               }
            }
         }

         // Get preferred cookies
         SAcookies cookies;
         cookies.bbuserid = -1; cookies.bbpassword = ""; cookies.browserName = "";

         if (useManualCook)
         {
            cookies.bbuserid = Convert.ToInt32(manualCookUser);
            cookies.bbpassword = manualCookPwd;
         }
         else
         {
            if (autoCookIE) { cookies = getCookiesIE(); }
            if (cookies.bbuserid == -1 && autoCookOpera) { cookies = getCookiesOpera(); }
            if (cookies.bbuserid == -1 && autoCookFF2) { cookies = getCookiesFF2(); }
            if (cookies.bbuserid == -1 && autoCookFF3) { cookies = getCookiesFF3(); }
         }

         if (cookies.bbuserid == -1)
         {
            MessageBox.Show("Couldn't find any valid Something Awful forum cookies", "LP Archiver");
            proceed = false;
         }

         if (!proceed)
            return;

         logFile = new StreamWriter(outDir + "archive.log");
         logger = new logForm();
         DateTime startTime = DateTime.Now;

         doLog("Started at " + startTime.ToString());
         doLog("");
         doLog("Using cookies from " + cookies.browserName);
         logger.Show();

         // Save entire thread to temporary file
         // (unless it's a VLP, in which case just the first page)
         if (! fetchThread(txtURL.Text, cookies, outDir + "fetchedThread.txt"))
         {
            MessageBox.Show("Download of thread failed - giving up", "LP Archiver");
            logFile.Close();
            return;
         }

         StreamReader inFile = new StreamReader(outDir + "fetchedThread.txt");
         string thisLine;
         int postCount = 0;

         // All this counting and chaptering is SSLP stuff. VLP only cares about the first post.
         if (optSSLP.Checked)
         {
            // Pre-count number of posts
            Regex postEndRE = new Regex("<!-- EndContentMarker --");

            doLog("* Pre-counting posts in thread");
            while ((thisLine = inFile.ReadLine()) != null)
               if (postEndRE.IsMatch(thisLine))
                  postCount++;
            doLog("- " + postCount + " found");

            // rewind
            inFile.BaseStream.Position = 0;
         } // end SSLP-only bit, seeing as I need the setup stuff just here for VLPs too

         // Ok, set up some general thread variables
         string threadAuthor = "?"; string postAuthor = "?";
         string firstPostDate = "?"; string lastPostDate = "?";
         string threadName = "?"; string threadURL = "?";

         getPostStuff postVarSet;
         postVarSet.threadFile = inFile;
         postVarSet.postNum = 0;
         postVarSet.amChapterTitleChecking = false;
         postVarSet.useAutoChapters = false;
         postVarSet.chapTitlesCount = 0;
         postVarSet.videoURLs = new ArrayList();
         postVarSet.videoTitles = new ArrayList();
         postVarSet.vidColumns = 1;
         postVarSet.imgCount = 0; postVarSet.imgSize = 0;
         postVarSet.vidSize = 0;
         postVarSet.smilies = new Hashtable();
         postVarSet.updCount = 0;
         postVarSet.gotUpdate = false;
         postVarSet.chapterTitles = new ArrayList();
         postVarSet.chapterTitles.Add("I am update 0. Nobody will ever see me :(");
         // I hear this 'SetCurrentDirectory' thing is popular lately
         postVarSet.outDirToo = outDir;
         postVarSet.postCountToo = postCount;
         postVarSet.errors = "";
         postVarSet.failedImages = new ArrayList();
         postVarSet.failedVideos = new ArrayList();
         postVarSet.postStats = new ArrayList();
         postVarSet.vlpContent = "";

         Regex authNameRE = new Regex(@"<dt class=""author(?: op)?"">(.+)</dt>");
         Regex postStartRE = new Regex(@"<!-- BeginContentMarker -->");
         Match lineM;

         // (VLP) not this bit, though
         if (optSSLP.Checked)
         {
            // See how many chapters have titles, so we can check the percentage to see if we want to use chapter titles...
            if (chkAutoChapterNaming.Checked)
            {
               doLog("* Pre-checking whether to use chapter titles");
               postVarSet.amChapterTitleChecking = true; postVarSet.useAutoChapters = true;

               while ((thisLine = inFile.ReadLine()) != null)
               {
                  lineM = authNameRE.Match(thisLine);
                  if (lineM.Success)
                  {
                     postAuthor = lineM.Groups[1].Value;
                     // Set the thread author as the author of the first post
                     if (threadAuthor == "?") { threadAuthor = postAuthor; }
                  }
                  else if (postStartRE.IsMatch(thisLine))
                  {
                     // Only do this if it's by the current author, or we're not using author posts only
                     // ** should be able to skip some of this with the thread fetcher only getting author posts anyway
                     if (postAuthor == threadAuthor || !chkAuthorPostsOnly.Checked)
                     {
                        getPost(ref postVarSet);
                     }
                  }
               }

               // Switch off auto chapters
               postVarSet.useAutoChapters = false;
               // unless the percentage of updates/updates with titles is met
               if ((postVarSet.chapTitlesCount / postVarSet.updCount) >= (Convert.ToInt32(txtAutoChapThresh.Text) / 100.0))
               {
                  postVarSet.useAutoChapters = true;
                  doLog("- Yes");
               }
               else { doLog(" - No"); }
               doLog("");

               // rewind and reset
               inFile.BaseStream.Position = 0;
               postVarSet.amChapterTitleChecking = false;
               postVarSet.updCount = 0; postVarSet.postNum = 0;
            }
         } // end SSLP-only bit

         // Go through the file until you hit a post or something else interesting, or the file ends
         while ((thisLine = inFile.ReadLine()) != null)
         {
            // Hey, look, it's a complete copy and paste from the chapter bit just up there ^^^
            lineM = authNameRE.Match(thisLine);
            if (lineM.Success)
            {
               postAuthor = lineM.Groups[1].Value;
               // Set the thread author as the author of the first post
               if (threadAuthor == "?") { threadAuthor = postAuthor; }
            }
            else if (postStartRE.IsMatch(thisLine))
            {
               // This shouldn't happen, but you never know, I guess
               if (optVLP.Checked && postVarSet.updCount != 0)
                  break;

               // Only do this if it's by the current author, or we're not using author posts only
               // ** should be able to skip some of this with the thread fetcher only getting author posts anyway
               if (postAuthor == threadAuthor || !chkAuthorPostsOnly.Checked)
               {
                  getPost(ref postVarSet);
               }
            }

            // gotUpdate is only set on if the last post retrieved was a valid update
            else if (postVarSet.gotUpdate && Regex.Match(thisLine, @"<td class=""postdate"">").Success)
            {
               // ugh
               thisLine = inFile.ReadLine(); // blank line
               thisLine = inFile.ReadLine(); // post number
               thisLine = inFile.ReadLine(); // thread/user numbers
               thisLine = inFile.ReadLine(); // This is the date that we actually want

               lastPostDate = Regex.Match(thisLine, @"(\w+ \d+, \d+) \d+:\d+").Groups[1].Value;
               if (firstPostDate == "?") { firstPostDate = lastPostDate; }

               // Got the first post (probably) and its date, so we're done for a VLP
               if (optVLP.Checked)
               {
                  // Last Update is not clearly defined for VLPs
                  lastPostDate = "?";
                  break;
               }

               // Tack date onto the end of the update in case of renumbering
               // (apparently?)
               // ,true == Append
               StreamWriter outFile = new StreamWriter(outDir + "Update " + postVarSet.updCount.ToString("00") + "/index.html", true);
               outFile.WriteLine("<!-- date " + lastPostDate + " -->");
               outFile.Close();

               postVarSet.gotUpdate = false;
            }

            // Thread title
            else if (Regex.Match(thisLine, "<title>").Success)
            {
               threadName = Regex.Match(thisLine, "<title>(.+) - The Something Awful Forums</title>").Groups[1].Value;
            }

            // Thread ID (for generating the correcter URL)
            // (this is actually still here even though search is down, just commented out)
            else if (Regex.Match(thisLine, @"<label>Search thread: ").Success)
            {
               threadURL = "http://forums.somethingawful.com/showthread.php?threadid=" +
                  Regex.Match(thisLine, @"name=""threadid"" value=""(\d+)""").Groups[1].Value;
            }
         }
         inFile.Close();

         if (optSSLP.Checked)
         {
            // Got the posts, just cleanup and finish
            // - specifically, delete the 'next' links from the final post
            string[] lastPost = File.ReadAllLines(outDir + "Update " + postVarSet.updCount.ToString("00") + "/index.html");
            StreamWriter newLastPost = new StreamWriter(outDir + "Update " + postVarSet.updCount.ToString("00") + "/index.html");
            bool checkingLines = false;
            foreach (string lPLine in lastPost)
            {
               if (checkingLines)
               {
                  if (Regex.IsMatch(lPLine, "Next Chapter"))
                     continue;
               }
               // If I don't do this and use WriteLine instead, it ends up with \r\n line endings
               newLastPost.Write(lPLine + "\n");

               if (Regex.IsMatch(lPLine, "<!-- begin (header|footer) -->"))
                  checkingLines = true;
               else if (Regex.IsMatch(lPLine, "<!-- end (header|footer) -->"))
                  checkingLines = false;
            }
            newLastPost.Close();
         }

         // Consolidated smiley-fetch
         if (postVarSet.smilies.Count > 0)
         {
            Directory.CreateDirectory(outDir + "Smilies");
            string smileyFile;
            WebClient downloader = new WebClient();

            foreach (string smileyURL in postVarSet.smilies.Keys)
            {
               smileyFile = (string)postVarSet.smilies[smileyURL];
               // Strip the "../Smilies/" from the start of the file
               smileyFile = Regex.Replace(smileyFile, "^.+/", "");

               // I've always been re-fetching these every time before this rewrite. That's weird.
               if (!File.Exists(outDir + "Smilies/" + smileyFile))
               {
                  doLog("* Fetching smiley " + smileyFile + " (" + smileyURL + ")");
                  if (chkDownloadFiles.Checked)
                  {
                     // This copy-paste shames me. More than usual. I would've made fetching
                     // an image a function if I'd thought about it. Or fetching a file
                     // in general for that matter...
                     try
                     {
                        downloader.DownloadFile(smileyURL, outDir + "Smilies/" + smileyFile);
                     }
                     catch (Exception ex)
                     {
                        doLog("X Download of smiley " + smileyFile + " (" + smileyURL + ") failed");
                        doLog("X " + ex.Message);
                     }
                  }
               }
            }
         }

         // Retry fetching failed images
         foreach (failedImage failImg in postVarSet.failedImages)
         {
            doLog("* Retrying download of " + failImg.imageInfo + " (" + failImg.imageURL + ")");

            // Hey, guess what's great? Pasting that image-fetch code in again
            WebClient downloader = new WebClient();
            bool dlSuccess = true;
            FileInfo testFile = null;

            // paintedover requires a forums referer
            if (failImg.imageURL.IndexOf("paintedover") > 0)
               downloader.Headers.Add("Referer", "http://forums.somethingawful.com/");

            try
            {
               downloader.DownloadFile(failImg.imageURL, failImg.imagePath);
               testFile = new FileInfo(failImg.imagePath);
            }
            catch (Exception ex)
            {
               doLog("X Still failed");
               doLog("X " + ex.Message);
               dlSuccess = false;
            }

            // update stats
            postVarSet.imgCount++; // I mean, technically there's an image in the page
            if (dlSuccess) { postVarSet.imgSize += testFile.Length; }
         }

         string addedDate = DateTime.Now.ToString("MMM dd, yyyy");

         // Create the main index
         doLog("* Creating main index page");
         StreamWriter mainIndex = new StreamWriter(outDir + "index.html");

         mainIndex.Write("" +
"<html>\n" +
"\n" +
"<head>\n" +
@"<link rel=""StyleSheet"" href=""" + cssURL + @""" type=""text/css"">" + "\n" +
"<TITLE>" + txtLPName.Text + " - Index</TITLE>\n" +
"</head>\n" +
"\n" +
"<body>\n" +
"<h1>Let's Play " + txtLPName.Text + "</h1>\n" +
"Author: " + threadAuthor + "<br>\n" +
"Collected By: " + txtCollName.Text + "<br>\n" +
@"Original Thread: <A HREF=""" + threadURL + @""">" + threadName + "</A><br>\n" +
"First Update: " + firstPostDate + "<br>\n" +
"Last Update: " + lastPostDate + "<br>\n" +
"Date Added: " + addedDate + "<br>\n" +
"\n" +
"<h1>Statistics</h1>\n");

         // I'm not sure whether or not this is impossible
         if (postVarSet.updCount == 0)
         {
            mainIndex.Write("No updates found!\n");
         }
         else
         {
            if (optSSLP.Checked)
            {
               mainIndex.Write("Number of Pictures: " + postVarSet.imgCount + "<br>\n");
               mainIndex.Write(String.Format("Total Size: {0:F2} MB<br>\n", (float)postVarSet.imgSize / 1048576));
               mainIndex.Write("Number of updates: " + postVarSet.updCount + "<br>\n");
               mainIndex.Write(String.Format("Pictures per Update: {0:F2}<br>\n", postVarSet.imgCount / postVarSet.updCount));
               mainIndex.Write(String.Format("Size per Update: {0:F2} MB<br>\n", (float)postVarSet.imgSize / postVarSet.updCount / 1048576));
            }
            else
            {
               int vCount = 0;
               foreach (string vURL in postVarSet.videoURLs)
               {
                  if (vURL != "" && vURL[0] != 'b') { vCount++; }
               }

               mainIndex.Write("Number of Videos: " + vCount);
            }
         }

         if (optSSLP.Checked)
         {
            mainIndex.Write("\n" +
"<br><br><br>\n" +
"<h1>Table of Contents</h1>\n" +
"<!-- begin TOC -->\n");

            for (int uNum = 1; uNum <= postVarSet.updCount; uNum++)
            {
               mainIndex.Write(@"<li><A HREF=""Update%20" + uNum.ToString("00") + @"/index.html"">" + postVarSet.chapterTitles[uNum] + "</A><br>\n");
            }

            mainIndex.Write("" +
"<!-- end TOC -->\n" +
"\n");
         }
         else
         {
            mainIndex.Write("\n<br><br><br>\n");
            mainIndex.Write("<h1>Introduction</h1>\n");
            mainIndex.Write(postVarSet.vlpContent);

            mainIndex.Write("" +
"<br><br><br>\n" +
"<h1>Videos</h1>" +
@"<TABLE border=""1"">" + "\n" +
@"<COL width=""825"">" + "\n");

            if (chkMultiVidsAreBackups.Checked)
            {
               int i = 0;
               string vTitle, vURL;
               while (i < postVarSet.videoURLs.Count)
               {
                  vTitle = (string)postVarSet.videoTitles[i];
                  mainIndex.Write("<TR><TD>" + vTitle);
                  i++;

                  while (i < postVarSet.videoURLs.Count)
                  {
                     vTitle = (string)postVarSet.videoTitles[i]; vURL = (string)postVarSet.videoURLs[i];
                     if (vURL == "") break;
                     vURL = vURL.TrimStart(new char[] { 'b' });

                     mainIndex.Write(@"<TD><a href=""" + vURL + @""" target=""_blank"" rel=""nofollow"">" +
                        vTitle + "</a>\n");
                     i++;
                  }
               }
            }
            else
            {
               for (int i = 0; i < postVarSet.videoURLs.Count; i++)
               {
                  mainIndex.Write(@"<TR><TD><a href=""" + postVarSet.videoURLs[i] + @""" target=""_blank"" rel=""nofollow"">" +
                     postVarSet.videoTitles[i] + "</a>\n");
               }
            }
            mainIndex.Write("</TABLE>\n");
         }

         mainIndex.Write("\n" +
"<br><br><br>\n" +
@"<A HREF=""/LetsPlay/"">Back to Let's Play Index</A>" + "\n" +
"</body>\n" +
"</html>\n");
         mainIndex.Close();

         doLog("");
         doLog("Summary statistics");
         int img, vid;
         string vTxt;
         for (int stNum = 0; stNum < postVarSet.postStats.Count; stNum++)
         {
            img = (int)(postVarSet.postStats[stNum]) / 100;
            vid = (int)(postVarSet.postStats[stNum]) % 100;
            if(img > 1) { vTxt = "s"; } else { vTxt = ""; }
            if(vid > 0) { vTxt += ", " + vid + " video"; if (vid > 1) { vTxt += "s"; } }
            doLog("Update " + (stNum + 1) + ": " + img + " image" + vTxt);
         }

         DateTime endTime = DateTime.Now;
         doLog("");
         doLog("Finished at " + endTime.ToString());
         TimeSpan elapsed = endTime - startTime;
         if (elapsed.TotalSeconds < 60)
            doLog("Elapsed time: " + elapsed.Seconds + " seconds");
         else
         {
            int eMin = elapsed.Minutes; int eSec = elapsed.Seconds;
            string secTxt, minTxt;
            secTxt = eSec + " second"; if (eSec != 1) { secTxt += "s"; }
            minTxt = eMin + " minute"; if (eMin != 1) { secTxt += "s"; }
            doLog("Elapsed time: " + minTxt + ", " + secTxt);
         }

         logFile.Close();
         MessageBox.Show("Fetch complete!", "LP Archiver");
      }

      private Stream getPage(WebRequest req)
      {
         // Get the Stream object for a single web page, with some error trapping

         HttpWebResponse threadFetcher;

         try
         {
            threadFetcher = (HttpWebResponse)req.GetResponse();
         }
         catch (Exception ex)
         {
            doLog("X Error occurred while fetching thread: " + ex.Message);
            return null;
         }

         if (threadFetcher.StatusCode != HttpStatusCode.OK)
         {
            doLog("X HTTP error received: " + threadFetcher.StatusDescription);
            return null;
         }

         return threadFetcher.GetResponseStream();
      }

      private bool fetchThread(string threadURL, SAcookies cookies, string outFilename)
      {
         if (File.Exists(outFilename))
         {
            if (MessageBox.Show("A fetched-thread output file already exists. Do you want to use this existing one?",
               "LP Archiver", MessageBoxButtons.YesNo) == DialogResult.Yes)
            {
               doLog("Using existing fetched-thread file"); doLog("");
               return true;
            }
         }

         doLog("* Fetching thread from " + threadURL);

         WebRequest threadReq = WebRequest.Create(threadURL);

         threadReq.Headers.Add("Cookie", "bbuserid=" + cookies.bbuserid +
            "; bbpassword=" + cookies.bbpassword);

         Stream respStream = getPage(threadReq);
         if(respStream == null) { return false; }

         StreamReader threadPage = new StreamReader(respStream);
         StreamWriter outFile = new StreamWriter(outFilename);

         Match lineM;
         Regex notLoggedInRE = new Regex(@"<div id=""notregistered"">"); bool notLoggedInChecking = true;

         // <div class="pages top">Pages (2): .............. </div>
         // <div class="pages top"> </div>
         Regex pgRE = new Regex(@"<div class=""pages top"">Pages \((\d+)\)");
         Regex pgRE1 = new Regex(@"<div class=""pages top""> </div>");

         Regex authRE = new Regex(@"&amp;userid=(\d+)"">\?</a>$");
         string thisLine;
         int numPages = 0; string authorID = null;

         while ((thisLine = threadPage.ReadLine()) != null)
         {
            lineM = pgRE.Match(thisLine);
            if (lineM.Success)
            {
               numPages = Convert.ToInt32(lineM.Groups[1].Value);
               // You must be past the point where that <div> appears if you've hit the page numbers
               notLoggedInChecking = false;
            }
            else if (pgRE1.Match(thisLine).Success)
            {
               numPages = 1;
               notLoggedInChecking = false;
            }

            // You appear not to be logged into the forums correctly
            if (notLoggedInChecking)
            {
               if (notLoggedInRE.IsMatch(thisLine))
               {
                  doLog("Not logged into the forums correctly");
                  if (MessageBox.Show("Your " + cookies.browserName + " cookies seem not to have logged you into the forums correctly. " +
                     "Some forum posts may appear incorrectly (eg word filters) or not at all. Go ahead anyway?",
                     "LP Archiver", MessageBoxButtons.YesNo) == DialogResult.No)
                  {
                     doLog("Giving up");
                     return false;
                  }
                  doLog("Carrying on anyway");
                  notLoggedInChecking = false;
               }
            }

            if (authorID == null)
            {
               lineM = authRE.Match(thisLine);
               if (lineM.Success)
                  authorID = lineM.Groups[1].Value;
            }

            if (! chkAuthorPostsOnly.Checked || optVLP.Checked)
            {
               // If it's author-only, and not a VLP, this first page wouldn't be much use
               outFile.WriteLine(thisLine);
            }
         }
         threadPage.Close();

         if (numPages == 0)
         {
            MessageBox.Show("No posts were found in this thread. It's likely that either the thread URL is wrong or you're not logged in correctly." +
               "You can check fetchedThread.txt for details.", "LP Archiver");
            doLog("No posts found, giving up");
            return false;
         }

         // Don't care about additional pages if it's a VLP
         if (! optVLP.Checked)
         {
            // If we're fetching pages only by the author, and we probably are,
            // then we have to do the page count again. Oh well.
            if (chkAuthorPostsOnly.Checked)
            {
               threadURL += "&userid=" + authorID;

               doLog("* Fetching author-only posts from " + threadURL);

               // Bleh. You can't just set a new URL on the existing WebRequest object
               threadReq = WebRequest.Create(threadURL);
               threadReq.Headers.Add("Cookie", "bbuserid=" + cookies.bbuserid +
                  "; bbpassword=" + cookies.bbpassword);

               respStream = getPage(threadReq);
               if (respStream == null) { return false; }

               threadPage = new StreamReader(respStream);
               numPages = 1;

               while ((thisLine = threadPage.ReadLine()) != null)
               {
                  lineM = pgRE.Match(thisLine);
                  if (lineM.Success)
                     numPages = Convert.ToInt32(lineM.Groups[1].Value);

                  outFile.WriteLine(thisLine);
               }
               threadPage.Close();
            }

            if (numPages > 1)
            {
               doLog("* Total of " + numPages + " pages to fetch");

               string pagelessURL = threadURL + "&pagenumber="; string pagefulURL;
               for (int i = 2; i <= numPages; i++)
               {
                  pagefulURL = pagelessURL + i;
                  doLog("* Fetching page from " + pagefulURL);

                  threadReq = WebRequest.Create(pagefulURL);
                  threadReq.Headers.Add("Cookie", "bbuserid=" + cookies.bbuserid +
                     "; bbpassword=" + cookies.bbpassword);

                  respStream = getPage(threadReq);
                  if (respStream == null) { return false; }

                  threadPage = new StreamReader(respStream);
                  while ((thisLine = threadPage.ReadLine()) != null) { outFile.WriteLine(thisLine); }
                  threadPage.Close();
               }
            }
         }

         outFile.Close();
         doLog("");

         return true;
      }

      // stuff is the best word ever
      private void getPost(ref getPostStuff stuff)
      {
         stuff.postNum++;

         int postImgNum = 0;
         ArrayList newVidTitles = new ArrayList(); ArrayList newVidURLs = new ArrayList();
         bool keepPost = false; bool inQuote = false;
         string chapterTitle = "";
         string postBody = "";
         Regex postEndRE = new Regex(@"<!-- EndContentMarker -->");
         Regex commentLineRE = new Regex(@"^\s*<!--.+?-->\s*$");
         // Why does the perl version have "<blockquote " for this? I'll keep that as a possible match just in case.
         Regex blockQuoteStartRE = new Regex(@"<blockquote");
         Regex blockQuoteEndRE = new Regex(@"</blockquote>");
         Hashtable images = new Hashtable();

         string thisLine;
         while ((thisLine = stuff.threadFile.ReadLine()) != null)
         {
            // End of post mark
            if (postEndRE.IsMatch(thisLine))
               break;

            // Toss lines that are just comments
            if (commentLineRE.IsMatch(thisLine))
               continue;

            // Starting/ending a quote
            if (blockQuoteStartRE.IsMatch(thisLine))
               inQuote = true;
            else if (blockQuoteEndRE.IsMatch(thisLine))
               inQuote = false;

            // Is this a (possible) chapter title?
            // First bolded text in post, or matches one of the Chapter Words
            if (!inQuote && stuff.useAutoChapters && chapterTitle == "")
            {
               bool isTitle = false;

               if (Regex.Match(thisLine, "<b>.+</b>").Success)
               {
                  chapterTitle = Regex.Match(thisLine, "<b>(.+)$").Groups[1].Value;
                  isTitle = true;
               }
               else
               {
                  // Matches a chapter word?
                  foreach (string cWord in chapterWords)
                  {
                     if(Regex.Match(thisLine, cWord).Success) {
                        isTitle = true;
                        chapterTitle = thisLine;
                        break;
                     }
                  }
               }

               if (isTitle)
               {
                  // Strip other tags
                  chapterTitle = Regex.Replace(chapterTitle, "</?[^>]+>", "");
               }
            }

            // Check for links to videos
            Regex aHrefRE = new Regex("<a href=");

            if (aHrefRE.IsMatch(thisLine))
            {
               bool isVideo;
               Match vidURLMatch;
               string lineTitle = "";

               // Store line title for video table row
               if (optVLP.Checked)
               {
                  lineTitle = thisLine;
                  // Strip tags (except images, in case of smileys)
                  lineTitle = Regex.Replace(lineTitle, "<img", "!img");
                  lineTitle = Regex.Replace(lineTitle, "</?[^>]+>", "");
                  lineTitle = Regex.Replace(lineTitle, "!img", "<img");
               }

               do
               {
                  string URL, title, siteName = "";

                  vidURLMatch = Regex.Match(thisLine, @"<a href=""(http://[^""]+)""[^>]*>([^<]+)</a>");
                  if (!vidURLMatch.Success)
                  {
                     // No title, or it's an image, or something. Must redo this better next time.
                     vidURLMatch = Regex.Match(thisLine, @"<a href=""(http://[^""]+)""");
                     URL = vidURLMatch.Groups[1].Value;
                     title = "(no title)";
                  }
                  else
                  {
                     URL = vidURLMatch.Groups[1].Value;
                     title = vidURLMatch.Groups[2].Value;
                  }

                  // Is it actually a video link?
                  isVideo = false;

                  for (int i = 0; i < videoSiteURLs.Length; i++)
                  {
                     if (Regex.Match(URL, videoSiteURLs[i]).Success)
                     {
                        isVideo = true;
                        siteName = videoSiteNames[i];
                        break;
                     }
                  }

                  if (isVideo)
                  {
                     if (optVLP.Checked)
                     {
                        // Do we already have this video?
                        int vidNumber = stuff.videoURLs.IndexOf(URL) + 1;
                        // Or as a backup video?
                        if (vidNumber == 0)
                           vidNumber = stuff.videoURLs.IndexOf("b" + URL) + 1;

                        // No? Then add it to the potential new video list (used if the post is kept)
                        if (vidNumber == 0)
                        {
                           string vTitle;

                           if (chkMultiVidsAreBackups.Checked)
                           {
                              // Tag secondary videos on the same line with a 'b' for backup
                              if (newVidURLs.Count > 0)
                                 URL = "b" + URL;

                              vTitle = siteName;
                           }
                           else
                           {
                              vTitle = lineTitle + " | " + siteName;
                           }

                           newVidTitles.Add(vTitle); newVidURLs.Add(URL);
                        }
                     }

                     // This is the only thing we care about if it's SSLP
                     if (!inQuote && chkKeepVideoOnlyPosts.Checked) { keepPost = true; }
                  }

                  // Avoid reprocessing this link
                  thisLine = aHrefRE.Replace(thisLine, "<x href=", 1);
               }
               while (aHrefRE.IsMatch(thisLine));

               // Put links back to normal
               thisLine = Regex.Replace(thisLine, "<x href=", "<a href=");

               if (newVidURLs.Count > 0)
               {
                  // Check for maximum possible number of columns
                  if (newVidURLs.Count > stuff.vidColumns)
                     stuff.vidColumns = newVidURLs.Count;

                  if (chkMultiVidsAreBackups.Checked)
                  {
                     newVidTitles.Insert(0, lineTitle); newVidURLs.Insert(0, "");
                  }

                  stuff.videoTitles.AddRange(newVidTitles); stuff.videoURLs.AddRange(newVidURLs);
                  newVidTitles.Clear(); newVidURLs.Clear();
               }
            }

            // Get images
            Regex imgRE1 = new Regex(@"<img (?:width=\d+ )?src=""(http://[^""]+)");
            Regex imgRE2 = new Regex(@"<a href=""(http://[^/]+waffleimages\.com[^""]+)");

            while (imgRE1.IsMatch(thisLine) || imgRE2.IsMatch(thisLine))
            {
               string imgURL, origImgURL, newImgLink;
               Match iMatch;

               if((iMatch = imgRE1.Match(thisLine)).Success) { imgURL = iMatch.Groups[1].Value; }
               else { imgURL = imgRE2.Match(thisLine).Groups[1].Value; }

               origImgURL = imgURL;

               // ImageSocket fix
               imgURL = Regex.Replace(imgURL, "http://imagesocket.com/", "http://content.imagesocket.com/");
               // SALR/other random #-adders stripping
               imgURL = Regex.Replace(imgURL, "#[^#]+$", "");
               // Un-mirror-server waffleimages links
               imgURL = Regex.Replace(imgURL, @"http://[^/]+.mirror.waffleimages.com/files/../([^\.]+)\.(.+)$", @"http://img.waffleimages.com/$1/waffle.$2");

               // Do we already have this (standard or smiley)?
               if (images.Contains(imgURL))
               {
                  // Yes - so link to the existing file
                  newImgLink = (string)images[imgURL];
               }
               else if (stuff.smilies.Contains(imgURL))
               {
                  newImgLink = (string)stuff.smilies[imgURL];
               }
               else if (Regex.Match(imgURL, "http://i.somethingawful.com/forumsystem/emoticons/").Success ||
                  Regex.Match(imgURL, "http://i.somethingawful.com/images/").Success ||
                  Regex.Match(imgURL, "http://fi.somethingawful.com/images/smilies/").Success)
               {
                  // Don't have it, but it's a smiley
                  string smileyFile = Regex.Match(imgURL, "/([^/]+)$").Groups[1].Value;
                  newImgLink = "../Smilies/" + smileyFile;
                  stuff.smilies[imgURL] = newImgLink;
               }
               else
               {
                  // Right, it's a standard image. Make up a new filename for it.
                  postImgNum++;

                  newImgLink = Regex.Match(imgURL, "/([^/]+)$").Groups[1].Value;
                  // I'm not sure why this matters particularly
                  if(! Regex.Match(newImgLink, @"\..{1,3}$").Success) { newImgLink += "."; }

                  // Clear out any odd characters
                  newImgLink = Regex.Replace(newImgLink, " ", "_");
                  newImgLink = Regex.Replace(newImgLink, @"%[\dA-F]{2}", "_");

                  // Add the image number
                  newImgLink = postImgNum + "-" + newImgLink;

                  // and store it in case it gets reused this post
                  images[imgURL] = newImgLink;

                  // Keep the post if we've exceeded the minimum required image count
                  // Well, unless it's a quoted image, then it doesn't count
                  if ((!inQuote) && (postImgNum >= Convert.ToInt32(txtMinImages.Text)))
                     keepPost = true;
               }

               // Modify the post with the new image link
               // I want an equivalent for perl's \Q...\E here for origImgURL, but
               // apparently there isn't one. This might be a problem. I really don't know.
               if(optSSLP.Checked)
                  thisLine = Regex.Replace(thisLine, origImgURL, newImgLink);
               else
                  // Also, VLPs images go in the Images subdirectory
                  thisLine = Regex.Replace(thisLine, origImgURL, "Images/" + newImgLink);
            }

            postBody += thisLine + "\n";
         }

         // OP is always kept regardless
         if (stuff.updCount == 0) { keepPost = true; }
         else if (txtMinImages.Text == "0")
         {
            // Special case. If min images is 0, keep everything
            keepPost = true;
         }

         // If we're just checking for chapter titles, increase a
         // couple of counts and then don't do the post-keeping stuff
         if (stuff.amChapterTitleChecking)
         {
            if (keepPost)
            {
               stuff.updCount++;
               if (chapterTitle != "") { stuff.chapTitlesCount++; }
               keepPost = false;
            }
         }

         if (!keepPost)
         {
            // Keeping requires: txtMinImages images, or a video href (if chkVideoPosts is on)
            // (which are not inside <blockquote>s)
            if (!stuff.amChapterTitleChecking)
               doLog("Skipping non-content post (would have been chapter " + (stuff.updCount + 1) + ")");
         }
         else
         {
            // Post is valid, so next time you see a post date, remember it
            stuff.gotUpdate = true;
            stuff.updCount++;

            doLog("* Creating update " + stuff.updCount);

            // Fix 'definitely' typo filter (this is probably pointless)
            postBody = Regex.Replace(postBody, @"\[NOTE: I AM TOO STUPID TO SPELL THE WORD ""DEFINITELY"" CORRECTLY\]", "definitely");
            // Not sure why I'm doing this at all
            postBody = Regex.Replace(postBody, "<br />", "<br>");

            // Auto-generate a chapter title if one wasn't found in the post
            if(chapterTitle == "")
               chapterTitle = "Chapter " + stuff.updCount;
            // and store it
            stuff.chapterTitles.Add(chapterTitle);

            if (optSSLP.Checked)
            {
               // Create directory
               Directory.CreateDirectory(stuff.outDirToo + "Update " + stuff.updCount.ToString("00"));
               // And generate the HTML file
               StreamWriter outFile = new StreamWriter(stuff.outDirToo + "Update " + stuff.updCount.ToString("00") + "/index.html");

               // I wish C# had 'here documents'
               outFile.Write("" +
"<html>\n" +
"\n" +
"<head>\n" +
@"<link rel=""StyleSheet"" href=""" + cssURL + @""" type=""text/css"">" + "\n" +
"<TITLE>" + txtLPName.Text + " - " + chapterTitle + "</TITLE>\n" +
"</head>\n" +
"\n" +
"<body>\n" +
"<br><br><br>\n");

               // Next/previous links
               outFile.WriteLine("<!-- begin header -->");
               addHeaderFooter(stuff.updCount, outFile);
               outFile.WriteLine("<!-- end header -->\n");
               outFile.WriteLine("<br><br><br>");
               // Main content
               outFile.Write(postBody);
               // footer
               outFile.WriteLine("<br><br><br>");
               outFile.WriteLine("<!-- begin footer -->");
               addHeaderFooter(stuff.updCount, outFile);
               outFile.WriteLine("<!-- end footer -->");

               outFile.WriteLine("</body>");
               outFile.WriteLine("</html>");
               outFile.Close();
            }
            else
            {
               // VLP just has a single Images directory
               Directory.CreateDirectory(stuff.outDirToo + "Images");
               stuff.vlpContent = postBody;
            }

            // Download images
            int iNum = 0; int iCount = images.Count;
            string imgPath, imgURL, endFetchMsg = "";
            FileInfo testFile;
            WebClient downloader;

            // Reverse the image hashtable so we can sort it by value, because apparently
            // there is no non-terrible way to do this

            // Wait, does this even matter? Who cares what order I download the images in?
            SortedList segami = new SortedList(new numericSorter());
            // Hey, I can't reuse imgURL here because foreach has a shitty syntax
            foreach (string tmpImgURL in images.Keys) { segami[images[tmpImgURL]] = tmpImgURL; }

            foreach (string imgFile in segami.Keys)
            {
               iNum++;
               if (optSSLP.Checked)
                  imgPath = stuff.outDirToo + "Update " + stuff.updCount.ToString("00") + "/" + imgFile;
               else
                  imgPath = stuff.outDirToo + "Images" + "/" + imgFile;

               imgURL = (string)segami[imgFile];

               // Skip already-downloaded files
               if((testFile = new FileInfo(imgPath)).Exists)
               {
                  doLog("Skipping image " + imgFile + " (" + imgURL + "), already exists");

                  // update stats anyway
                  stuff.imgCount++;
                  stuff.imgSize += testFile.Length;

                  // onto the next image
                  continue;
               }
               // this is the part where I re-get the last image that
               // was skipped in case it was cut off mid-download
               // ...or it would be if I could be bothered to do that bit

               // Do the download
               if(optSSLP.Checked)
                  endFetchMsg = " for post " + stuff.postNum + "/" + stuff.postCountToo;

               doLog("* Fetching image " + iNum + "/" + iCount + " (" + imgURL + ")" + endFetchMsg);

               downloader = new WebClient();

               // paintedover requires a forums referer
               if (imgURL.IndexOf("paintedover") > 0)
                  downloader.Headers.Add("Referer", "http://forums.somethingawful.com/");

               if (chkDownloadFiles.Checked)
               {
                  bool dlSuccess = true;

                  try
                  {
                     downloader.DownloadFile(imgURL, imgPath);
                     testFile = new FileInfo(imgPath);
                  }
                  catch(Exception e)
                  {
                     doLog("X Download of image " + imgFile + " (" + imgURL + ") failed");
                     doLog("X " + e.Message);
                     dlSuccess = false;
                  }

                  if (dlSuccess)
                  {
                     // update stats
                     stuff.imgCount++;
                     stuff.imgSize += testFile.Length;
                  }
                  else
                  {
                     failedImage fI = new failedImage();
                     fI.imagePath = imgPath; fI.imageURL = imgURL;
                     fI.imageInfo = "image " + iNum + " in update " + stuff.updCount;
                     stuff.failedImages.Add(fI);
                  }
               }
               // if not downloading anyway, just increase the image count
               else { stuff.imgCount++; }
            }
         }
      }

      void addHeaderFooter(int updateNum, StreamWriter outFile)
      {
         if(updateNum > 1)
            outFile.WriteLine(@"<A HREF=""../Update%20" + (updateNum - 1).ToString("00") + @"/index.html"">&lt;&lt; Previous Chapter</A><br>");

         outFile.WriteLine(@"<A HREF=""../Update%20" + (updateNum + 1).ToString("00") + @"/index.html"">&gt;&gt; Next Chapter</A><br>");
         outFile.WriteLine(@"<A HREF=""../index.html"">^^ Index</A><br>");
      }


      public class numericSorter : IComparer
      {
         public int Compare(object x, object y)
         {
            // So the strings are like '1-blah.jpg' and '15-bloop.jpg'
            string xx = (string)x; string yy = (string)y;

            return Convert.ToInt32(xx.Substring(0, xx.IndexOf('-'))) - Convert.ToInt32(yy.Substring(0, yy.IndexOf('-')));
         }
      }
   }
}